from typing import Protocol, Tuple, Any, List

from seal.model.result import Result, Results


class IExecutor(Protocol):

    def find(self, sql: str, args: Tuple[Any, ...], bean_type: Any) -> Result:
        ...

    def find_list(self, sql: str, args: Tuple[Any, ...], bean_type: Any) -> Results:
        ...

    def count(self, sql: str, args: Tuple[Any, ...]) -> int | None:
        ...

    def update(self, sql: str, args: Tuple[Any, ...]) -> int | None:
        ...

    def insert(self, sql: str, args: Tuple[Any, ...]) -> int | None:
        ...

    def insert_bulk(self, sql: str, args: List[Tuple[Any, ...]]) -> int | None:
        ...

    def custom_query(self, sql: str, args=Tuple[Any, ...]) -> Results:
        ...

    def custom_update(self, sql: str, args=Tuple[Any, ...]) -> int | None:
        ...



from .data_source_protocol import IDataSource
from .database_connection_protocol import IDatabaseConnection
from .executor_protocol import IExecutor

__all__ = ['IExecutor', 'IDataSource', 'IDatabaseConnection']



from .condition import ConditionTree, Condition


class Wrapper:
    def __init__(self,
                 tenant_field=None,
                 tenant_value=None,
                 logical_deleted_field=None,
                 logical_deleted_value_true=None,
                 logical_deleted_value_false=None):
        self.condition_tree = ConditionTree()
        self.tenant_field = tenant_field
        self.tenant_value = tenant_value
        self.logical_deleted_field = logical_deleted_field
        self.logical_deleted_value_true = logical_deleted_value_true
        self.logical_deleted_value_false = logical_deleted_value_false

    def eq(self, field, value) -> 'Wrapper':
        self.condition_tree.add_condition(Condition(field, value))
        return self

    def ne(self, field, value) -> 'Wrapper':
        self.condition_tree.add_condition(Condition(field, value, '!='))
        return self

    def gt(self, field, value) -> 'Wrapper':
        self.condition_tree.add_condition(Condition(field, value, '>'))
        return self

    def ge(self, field, value) -> 'Wrapper':
        self.condition_tree.add_condition(Condition(field, value, '>='))
        return self

    def lt(self, field, value) -> 'Wrapper':
        self.condition_tree.add_condition(Condition(field, value, '<'))
        return self

    def le(self, field, value) -> 'Wrapper':
        self.condition_tree.add_condition(Condition(field, value, '<='))
        return self

    def in_(self, field, value) -> 'Wrapper':
        self.condition_tree.add_condition(Condition(field, value, 'in'))
        return self

    def l_like(self, field, value) -> 'Wrapper':
        self.condition_tree.add_condition(Condition(field, f'%{value}', 'like'))
        return self

    def r_like(self, field, value) -> 'Wrapper':
        self.condition_tree.add_condition(Condition(field, f'{value}%', 'like'))
        return self

    def like(self, field, value) -> 'Wrapper':
        self.condition_tree.add_condition(Condition(field, f'%{value}%', 'like'))
        return self

    def or_(self, wrapper: 'Wrapper') -> 'Wrapper':
        wrapper.condition_tree.or_()
        self.condition_tree.add_tree(wrapper.condition_tree)
        return self

    def handle_public_fields(self, **options):
        if 'logical_deleted' in options:
            self.eq(options['logical_deleted'], 0)
            return
        if self.tenant_field is not None:
            if self.tenant_value is None:
                raise ValueError('tenant value not set')
            self.eq(self.tenant_field, self.tenant_value)

        if self.logical_deleted_field is not None:
            if self.logical_deleted_value_true is None or self.logical_deleted_value_false is None:
                raise ValueError('logical deleted field and value not set')
            self.eq(self.logical_deleted_field, self.logical_deleted_value_false)



import uuid
from contextvars import ContextVar

from loguru import logger

from seal.db.protocol import IDataSource
from seal.db.protocol import IDatabaseConnection


class SqlContext:
    def __init__(self):
        self._tx: IDatabaseConnection | None = None
        self._tx_id = None

    def begin(self, ds: IDataSource):
        self._tx_id = uuid.uuid4()
        self._tx = ds.get_connection()
        self._tx.begin()

        logger.debug(f'begin transaction: {self._tx_id}')

    def commit(self):
        self._tx.commit()
        logger.debug(f'commit transaction: {self._tx_id}')
        self._tx.close()

    def rollback(self):
        self._tx.rollback()
        logger.debug(f'rollback transaction: {self._tx_id}')
        self._tx.close()

    def tx(self):
        return self._tx

    def tx_id(self):
        return self._tx_id


sql_context = ContextVar('sql_context', default=SqlContext())
